# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class MachineGroupTerminateReason(schemas.EnumBase, schemas.StrSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    An enumeration.
    """

    class MetaOapg:
        enum_value_to_name = {
            "user_request": "USER_REQUEST",
            "ttl_exceeded": "TTL_EXCEEDED",
            "idle_timeout": "IDLE_TIMEOUT",
            "credits_exhausted": "CREDITS_EXHAUSTED",
        }

    @schemas.classproperty
    def USER_REQUEST(cls):
        return cls("user_request")

    @schemas.classproperty
    def TTL_EXCEEDED(cls):
        return cls("ttl_exceeded")

    @schemas.classproperty
    def IDLE_TIMEOUT(cls):
        return cls("idle_timeout")

    @schemas.classproperty
    def CREDITS_EXHAUSTED(cls):
        return cls("credits_exhausted")



"""Remove the user's remote storage contents via CLI."""
import argparse
import sys

import inductiva
from inductiva.tasks.methods import get_all
from ...localization import translator as __
from inductiva.utils.input_functions import user_confirmation_prompt


def remove(args):
    """Remove user's remote storage contents."""
    is_all = args.all
    task_ids = args.id
    confirm = args.confirm

    if is_all and task_ids:
        print(
            "inductiva storage remove: error: "
            "argument id(s) not allowed with argument --all/-a",
            file=sys.stderr)
        return 1
    if not is_all and not task_ids:
        print(
            "inductiva storage remove: error: "
            "argument id(s) or --all/-a required",
            file=sys.stderr)
        return 1

    task_ids = set(task_ids)

    if is_all:
        all_tasks = get_all()
        task_ids = [task.id for task in all_tasks]

    if not confirm:
        confirm = user_confirmation_prompt(
            task_ids, __("storage-prompt-remove-all"),
            __("storage-prompt-remove-big", len(task_ids)),
            __("storage-prompt-remove-small"), is_all)

    if not confirm:
        return 0

    failed = False
    for task_id in task_ids:
        if not inductiva.tasks.Task(task_id).remove_remote_files(verbose=False):
            print(f"Failed to remove the following task storage: {task_id}\n",
                  file=sys.stderr)
            failed = True

    if failed:
        return 1

    print("All tasks storage removed successfully.")
    return 0


def register(parser):
    subparser = parser.add_parser("remove",
                                  aliases=["rm"],
                                  help="Remove remote storage entries.",
                                  formatter_class=argparse.RawTextHelpFormatter)
    subparser.description = (
        "The `inductiva storage remove` command deletes specified data"
        " from the platform.\n"
        "It targets a specific task files for removal. Use with caution as "
        "this action is irreversible.\n\n")
    subparser.add_argument("id",
                           type=str,
                           nargs="*",
                           help="Id(s) of the task(s) to remove from remote "
                           "storage.")

    subparser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        dest="confirm",
        default=False,
        help="Sets any confirmation values to \"yes\" "
        "automatically. Users will not be asked for "
        "confirmation to remove path(s) from remote storage.")

    subparser.add_argument("-a",
                           "--all",
                           action="store_true",
                           default=False,
                           help="Remove all tasks from remote storage.")

    subparser.set_defaults(func=remove)



#pylint: disable=missing-module-docstring
from .machines import MachineGroup, ElasticMachineGroup
from .machine_types import get_available_machine_types
from .machine_groups import estimate_machine_cost
from .machine_cluster import MPICluster
from .machine_types import list_available_machines
from .machine_groups import get



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class Project(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "num_tasks",
            "name",
            "created_at",
            "id",
        }

        class properties:

            class name(schemas.StrSchema):
                pass

            id = schemas.UUIDSchema
            created_at = schemas.DateTimeSchema
            num_tasks = schemas.IntSchema

            class task_status_overview(schemas.DictSchema):

                class MetaOapg:
                    additional_properties = schemas.IntSchema

                def __getitem__(
                        self, name: typing.Union[
                            str,
                        ]) -> MetaOapg.additional_properties:
                    # dict_instance[name] accessor
                    return super().__getitem__(name)

                def get_item_oapg(
                        self, name: typing.Union[
                            str,
                        ]) -> MetaOapg.additional_properties:
                    return super().get_item_oapg(name)

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[
                        MetaOapg.additional_properties,
                        decimal.Decimal,
                        int,
                    ],
                ) -> 'task_status_overview':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "name": name,
                "id": id,
                "created_at": created_at,
                "num_tasks": num_tasks,
                "task_status_overview": task_status_overview,
            }

    num_tasks: MetaOapg.properties.num_tasks
    name: MetaOapg.properties.name
    created_at: MetaOapg.properties.created_at
    id: MetaOapg.properties.id

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["created_at"]
    ) -> MetaOapg.properties.created_at:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["num_tasks"]
    ) -> MetaOapg.properties.num_tasks:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["task_status_overview"]
    ) -> MetaOapg.properties.task_status_overview:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "name",
        "id",
        "created_at",
        "num_tasks",
        "task_status_overview",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["created_at"]
    ) -> MetaOapg.properties.created_at:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["num_tasks"]
    ) -> MetaOapg.properties.num_tasks:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["task_status_overview"]
    ) -> typing.Union[MetaOapg.properties.task_status_overview, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "name",
        "id",
        "created_at",
        "num_tasks",
        "task_status_overview",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        num_tasks: typing.Union[
            MetaOapg.properties.num_tasks,
            decimal.Decimal,
            int,
        ],
        name: typing.Union[
            MetaOapg.properties.name,
            str,
        ],
        created_at: typing.Union[
            MetaOapg.properties.created_at,
            str,
            datetime,
        ],
        id: typing.Union[
            MetaOapg.properties.id,
            str,
            uuid.UUID,
        ],
        task_status_overview: typing.Union[
            MetaOapg.properties.task_status_overview, dict,
            frozendict.frozendict, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'Project':
        return super().__new__(
            cls,
            *_args,
            num_tasks=num_tasks,
            name=name,
            created_at=created_at,
            id=id,
            task_status_overview=task_status_overview,
            _configuration=_configuration,
            **kwargs,
        )



"""Module for consuming the streams of a task through a websocket.

Task logs are output in the following format:

$ induction logs <task_id>
<log line 1>                                        log content
<log line 2>
  ...
<log line N>
                                                    `SPACING` black lines
 ● 0:00:02 Streaming output from task <task_id>     connection status footer
"""
from threading import Thread, RLock
from urllib.parse import urlencode
from datetime import timedelta
from typing import IO
import logging
import sched
import json
import sys
import time
import os

import websocket

import inductiva
from inductiva import constants

logger = logging.getLogger("websocket")
logger.setLevel(logging.CRITICAL)

SPACING = 1
ESC = "\u001B["
RESET = f"{ESC}0m"
EMPHASIS = f"{ESC}2m"
FAILURE = f"{ESC}31m"
SUCCESS = f"{ESC}32m"
PENDING = f"{ESC}33m"
CLEAR_LINE = f"{ESC}0K"
UP = f"{ESC}{SPACING}A"
DOWN = f"{ESC}{SPACING}B"

ANSI_ENABLED = os.getenv("ANSI_ENABLED", "1") == "1"


class TaskStreamConsumer:
    """Consumer of the streams of a running task through a websocket.

    Attributes:
        RECONNECT_DELAY_SEC (float): Delay interval when reconnecting in
            seconds.
        PING_INTERVAL_SEC (float): Time interval, in seconds, to automatically
            send a “ping” command  to the websocket (in seconds).
            If set to 0, no ping is sent periodically.
        PING_TIMEOUT_SEC (float): Timeout (in seconds) to close the stream
            if the pong message is not received.
    """
    RECONNECT_DELAY_SEC = 5.
    PING_INTERVAL_SEC = 15.
    PING_TIMEOUT_SEC = 5.
    TICK_INTERVAL_SEC = 1.
    END_OF_STREAM = "<<end_of_stream>>"
    ALLOWED_IO_STREAMS = {
        "std_err": "standard error",
        "std_out": "standard output"
    }

    def __init__(self,
                 task_id: str,
                 fout: IO = sys.stdout,
                 ferr: IO = sys.stderr,
                 io_stream: str = None,
                 no_color: bool = False):
        """Initialize websocket connection to the task STDOUT & STDERR streams.

        Args:
            task_id (int): ID of the task for which to get the STDOUT & STDERR
                streams.
            fout (IO): I/O streams, such as returned by open(), for the STDOUT.
            ferr (IO): I/O streams, such as returned by open(), for the STDERR.
            io_stream (str): The type of stream to consume. It can be either
                "std_out" or "std_err". If not provided, both are consumed.
            no_color (bool): Disable ANSI color codes in the output. Default is
                False.
        """
        self.task_id = task_id
        self.fout = fout
        self.ferr = ferr

        # Check if the file handles are interactive and if ANSI is enabled.
        # Note that there is no point in using ANSI if ferr is not interactive
        self._ferr_isatty = ANSI_ENABLED and self.ferr.isatty()
        self._fout_isatty = ANSI_ENABLED and self.fout.isatty()
        self._fout_isatty &= self._ferr_isatty
        self._message_formatter = self._get_message_formatter()
        self._status_formatter = self._get_status_formatter()

        # housekeeping variables
        self._conn_start_time = None
        self._conn_alive_sec = 0.0
        self._conn_opened = False
        self._prev_status = None
        self._conn_count = 0
        self._lock = RLock()

        self._last_message_timestamp = None
        self._keyboard_interrupt = False
        self._end_of_stream = False

        if io_stream and io_stream not in self.ALLOWED_IO_STREAMS:
            raise ValueError(
                f"Invalid io_stream: {io_stream}\n"
                f"Allowed values are: {list(self.ALLOWED_IO_STREAMS.keys())}.")
        self.io_stream = io_stream
        self.no_color = no_color

    def __on_message(self, ws: websocket.WebSocketApp, message: str):
        data = json.loads(message)
        streams = data.get("streams")
        for stream in streams:
            self._last_message_timestamp, msg = stream["values"][0]

            if msg == self.END_OF_STREAM:
                self._end_of_stream = True
                ws.close()
                return
            # Only writes in red if we are receiving stdout and stderr
            # and the output is stderr
            is_stderr = (self.io_stream is None and
                         stream["stream"]["io_type"] == "IOTypes.STD_ERR")
            self._write_message(msg.rstrip(),
                                is_stderr=is_stderr and not self.no_color)

    def _get_message_formatter(self):
        """Get the message formatter based on the output file
        handle interactivity.
        """
        if self._fout_isatty:
            return self._ansi_message_formatter
        return self._plain_message_formatter

    def _get_status_formatter(self):
        """Get the message formatter based on the error file
        handle interactivity.
        """
        if self._ferr_isatty:
            return self._ansi_status_formatter
        return self._plain_status_formatter

    def _ansi_message_formatter(self, msg, is_stderr=False):
        """Fancy message formatter for formatted message output using
        ANSI escape codes.
        """
        # 1. move up SPACING lines
        # 2. move to beginning of line and clear line
        # 3. write message
        # 4. add new line
        # 5. move down SPACING lines
        # 6. move to beginning of line and clear line
        # 7. add new blank line
        if is_stderr:
            m = f"{UP}\r{CLEAR_LINE}{FAILURE}{msg}{RESET}{DOWN}\r{CLEAR_LINE}\n"
        else:
            m = f"{UP}\r{CLEAR_LINE}{msg}{DOWN}\r{CLEAR_LINE}\n"
        return m

    def _plain_message_formatter(self, msg, **_):
        """Simple message formatter for plain message output."""
        return msg + "\n"

    def _ansi_status_formatter(self, msg, status):
        """Fancy message formatter for formatted status message output using
        ANSI escape codes.
        """
        if self._conn_start_time is None:
            elapsed = ""
        else:
            alive = timedelta(seconds=int(self._conn_alive_sec))
            elapsed = f"{str(alive)} "

        # 1. go to beginning of line
        # 2. clear line
        # 3. set color for bullet
        #   3.1 write bullet icon
        #   3.2 reset color
        # 4. set emphasis formatter
        #   4.1. write timer message (if any)
        #   4.2. write status message
        # 5. reset everything
        s = f"\r{CLEAR_LINE}{status} ● {RESET}{EMPHASIS}{elapsed}{msg}{RESET}"
        return s

    # _ to make it have the same number or arguments as the other formatters
    # In this case, the second argument is not used.
    def _plain_status_formatter(self, msg, _):
        """Simple message formatter for plain status message output."""
        return msg + "\n"

    def _write_message(self, msg, is_stderr=False):
        """Write the message to the output file handle

        Args:
            msg (str): message to be written to output.
        """
        s = self._message_formatter(msg, is_stderr=is_stderr)
        self.fout.write(s)
        self.fout.flush()
        self._redraw_status()

    def _write_status(self, msg, status):
        """Write the status message to the error file handle

        Args:
            msg (str): Status message.
            status (str): Status formatting code.
        """
        s = self._status_formatter(msg, status)
        self.ferr.write(s)
        self.ferr.flush()

    def _update_status(self, msg, status):
        """Update the connection status message.

        Args:
            msg (str): Status message.
            status (str): Status formatting code.
        """
        with self._lock:
            self._prev_status = (msg, status)
            self._write_status(msg, status)

    def _redraw_status(self):
        """Redraw the connection status message."""
        if self._prev_status and self._ferr_isatty:
            self._update_status(*self._prev_status)

    def __on_error(self, ws: websocket.WebSocketApp, error: Exception):
        """Callback for websocket error event."""
        if isinstance(error, KeyboardInterrupt):
            self._keyboard_interrupt = True
            ws.close()
        else:
            self._update_status(str(error), FAILURE)
            time.sleep(2)

    def __on_close(self, unused_ws, close_status_code, close_message):
        """Callback for websocket close event."""

        if close_status_code or close_message:
            close_msg = f" (code: {close_status_code}, msg: {close_message})"
        else:
            close_msg = ""

        if self._conn_opened:
            if self._end_of_stream:
                msg = f"Received end of stream from task {self.task_id}"
                status = SUCCESS
            else:
                msg = f"Disconnected from task {self.task_id}"
                status = SUCCESS if self._keyboard_interrupt else FAILURE
            msg = f"{msg}{close_msg}."
        else:
            msg = f"Failed to connect to task {self.task_id}{close_msg}."
            status = FAILURE

        self._update_status(msg, status)

    def __on_open(self, unused_ws):
        """Callback for websocket open event."""
        self._conn_start_time = time.time()
        self._conn_opened = True
        self._conn_count += 1

        output = self.ALLOWED_IO_STREAMS.get(self.io_stream, "output")
        msg = f"Streaming {output} from task {self.task_id}"
        if ANSI_ENABLED and not self._fout_isatty:
            msg += " (redirected)"

        self._update_status(msg, SUCCESS)

    def __init_status(self):
        """Allocate sufficient lines for the status message."""
        if self._ferr_isatty:
            self.ferr.write(SPACING * "\n")
            self.ferr.flush()

    def _run_websocket(self):
        """Run the websocket application.
        If connection is lost, reconnect after a delay
        and update the start query to get the logs from the last
        received timestamp."""

        endpoint = constants.LOGS_WEBSOCKET_URL + "/loki/api/v1/tail?"
        if self.io_stream:
            io_type = f'io_type="IOTypes.{self.io_stream.upper()}"'
            params = {
                "query": f'{{{io_type},task_id="{self.task_id}"}}',
                "limit": 500
            }
        else:
            params = {"query": f'{{task_id="{self.task_id}"}}', "limit": 500}
        while True:
            url = endpoint + urlencode(params)
            ws = websocket.WebSocketApp(
                url,
                header={"X-API-Key": inductiva.get_api_key()},
                on_open=self.__on_open,
                on_error=self.__on_error,
                on_close=self.__on_close,
                on_message=self.__on_message)

            ws.run_forever(reconnect=0,
                           ping_interval=self.PING_INTERVAL_SEC,
                           ping_timeout=self.PING_TIMEOUT_SEC)

            if self._end_of_stream or self._keyboard_interrupt:
                break

            if self._last_message_timestamp is not None:
                params["start"] = int(self._last_message_timestamp) + 1
                params["limit"] = 10000

            time.sleep(self.RECONNECT_DELAY_SEC)

    def run_forever(self):
        """Stream the STDOUT & STDERR of a task through the websocket."""

        msg = f"Opening connection to task {self.task_id}..."
        self.__init_status()
        self._update_status(msg, PENDING)
        time.sleep(2)

        if scheduler := self._get_ontick_scheduler():
            thread = Thread(target=scheduler.run)
            thread.daemon = True
            thread.start()

        self._run_websocket()

    def _get_ontick_scheduler(self):
        """Get a scheduler for the connection alive ticker.

        Gets a scheduler that runs every second to update the time elapsed
        since the connection was opened.
        """
        if not self._ferr_isatty:
            return None

        scheduler = sched.scheduler(time.time, time.sleep)
        scheduler.enter(self.TICK_INTERVAL_SEC, 1, self._ontick, (scheduler,))
        return scheduler

    def _ontick(self, scheduler):
        """Update the connection alive time every second.

        Updates the time elapsed since the connection was opened, redraws
        the status message and schedules the next update.
        """
        if self._conn_start_time is not None:
            self._conn_alive_sec = time.time() - self._conn_start_time
            self._redraw_status()

        # schedule the next update
        scheduler.enter(self.TICK_INTERVAL_SEC, 1, self._ontick, (scheduler,))



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class MachineInfo(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "cpu_count_logical",
            "cpu_count_physical",
            "memory_bytes",
            "uuid",
        }

        class properties:
            uuid = schemas.StrSchema
            cpu_count_logical = schemas.IntSchema
            cpu_count_physical = schemas.IntSchema
            memory_bytes = schemas.IntSchema

            class host_name(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'host_name':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class started_at(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'started_at':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class terminated_at(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'terminated_at':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class last_seen_at(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'last_seen_at':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class current_task_id(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'current_task_id':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "uuid": uuid,
                "cpu_count_logical": cpu_count_logical,
                "cpu_count_physical": cpu_count_physical,
                "memory_bytes": memory_bytes,
                "host_name": host_name,
                "started_at": started_at,
                "terminated_at": terminated_at,
                "last_seen_at": last_seen_at,
                "current_task_id": current_task_id,
            }

    cpu_count_logical: MetaOapg.properties.cpu_count_logical
    cpu_count_physical: MetaOapg.properties.cpu_count_physical
    memory_bytes: MetaOapg.properties.memory_bytes
    uuid: MetaOapg.properties.uuid

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["uuid"]
    ) -> MetaOapg.properties.uuid:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["cpu_count_logical"]
    ) -> MetaOapg.properties.cpu_count_logical:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["cpu_count_physical"]
    ) -> MetaOapg.properties.cpu_count_physical:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["memory_bytes"]
    ) -> MetaOapg.properties.memory_bytes:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["host_name"]
    ) -> MetaOapg.properties.host_name:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["started_at"]
    ) -> MetaOapg.properties.started_at:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["terminated_at"]
    ) -> MetaOapg.properties.terminated_at:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["last_seen_at"]
    ) -> MetaOapg.properties.last_seen_at:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["current_task_id"]
    ) -> MetaOapg.properties.current_task_id:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "uuid",
        "cpu_count_logical",
        "cpu_count_physical",
        "memory_bytes",
        "host_name",
        "started_at",
        "terminated_at",
        "last_seen_at",
        "current_task_id",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["uuid"]
    ) -> MetaOapg.properties.uuid:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["cpu_count_logical"]
    ) -> MetaOapg.properties.cpu_count_logical:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["cpu_count_physical"]
    ) -> MetaOapg.properties.cpu_count_physical:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["memory_bytes"]
    ) -> MetaOapg.properties.memory_bytes:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["host_name"]
    ) -> typing.Union[MetaOapg.properties.host_name, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["started_at"]
    ) -> typing.Union[MetaOapg.properties.started_at, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["terminated_at"]
    ) -> typing.Union[MetaOapg.properties.terminated_at, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["last_seen_at"]
    ) -> typing.Union[MetaOapg.properties.last_seen_at, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["current_task_id"]
    ) -> typing.Union[MetaOapg.properties.current_task_id, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "uuid",
        "cpu_count_logical",
        "cpu_count_physical",
        "memory_bytes",
        "host_name",
        "started_at",
        "terminated_at",
        "last_seen_at",
        "current_task_id",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        cpu_count_logical: typing.Union[
            MetaOapg.properties.cpu_count_logical,
            decimal.Decimal,
            int,
        ],
        cpu_count_physical: typing.Union[
            MetaOapg.properties.cpu_count_physical,
            decimal.Decimal,
            int,
        ],
        memory_bytes: typing.Union[
            MetaOapg.properties.memory_bytes,
            decimal.Decimal,
            int,
        ],
        uuid: typing.Union[
            MetaOapg.properties.uuid,
            str,
        ],
        host_name: typing.Union[MetaOapg.properties.host_name, dict,
                                frozendict.frozendict, str, date, datetime,
                                uuid.UUID, int, float, decimal.Decimal, bool,
                                None, list, tuple, bytes, io.FileIO,
                                io.BufferedReader,
                                schemas.Unset] = schemas.unset,
        started_at: typing.Union[MetaOapg.properties.started_at, dict,
                                 frozendict.frozendict, str, date, datetime,
                                 uuid.UUID, int, float, decimal.Decimal, bool,
                                 None, list, tuple, bytes, io.FileIO,
                                 io.BufferedReader,
                                 schemas.Unset] = schemas.unset,
        terminated_at: typing.Union[MetaOapg.properties.terminated_at, dict,
                                    frozendict.frozendict, str, date, datetime,
                                    uuid.UUID, int, float, decimal.Decimal,
                                    bool, None, list, tuple, bytes, io.FileIO,
                                    io.BufferedReader,
                                    schemas.Unset] = schemas.unset,
        last_seen_at: typing.Union[MetaOapg.properties.last_seen_at, dict,
                                   frozendict.frozendict, str, date, datetime,
                                   uuid.UUID, int, float, decimal.Decimal, bool,
                                   None, list, tuple, bytes, io.FileIO,
                                   io.BufferedReader,
                                   schemas.Unset] = schemas.unset,
        current_task_id: typing.Union[MetaOapg.properties.current_task_id, dict,
                                      frozendict.frozendict, str, date,
                                      datetime, uuid.UUID, int, float,
                                      decimal.Decimal, bool, None, list, tuple,
                                      bytes, io.FileIO, io.BufferedReader,
                                      schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'MachineInfo':
        return super().__new__(
            cls,
            *_args,
            cpu_count_logical=cpu_count_logical,
            cpu_count_physical=cpu_count_physical,
            memory_bytes=memory_bytes,
            uuid=uuid,
            host_name=host_name,
            started_at=started_at,
            terminated_at=terminated_at,
            last_seen_at=last_seen_at,
            current_task_id=current_task_id,
            _configuration=_configuration,
            **kwargs,
        )



"""Methods that interact with the lower-level inductiva-web-api-client.

The relevant function for usage outside of this file is invoke_async_api().
Check the demos directory for examples on how it is used.
"""
import os
import time
import tqdm
import tqdm.utils
import signal
import urllib3
import decimal
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Type

import logging

import inductiva
from inductiva.client import ApiClient, ApiException, Configuration
from inductiva.client.apis.tags.tasks_api import TasksApi
from inductiva.client.models import TaskRequest, TaskStatus, TaskSubmittedInfo
from inductiva import types, constants
from inductiva.utils.data import (get_validate_request_params, pack_input)
from inductiva.utils import format_utils, files


def get_api_config() -> Configuration:
    """Returns an API configuration object."""
    api_key = inductiva.get_api_key()

    api_config = Configuration(host=inductiva.api_url)
    api_config.api_key["APIKeyHeader"] = api_key

    return api_config


def get_client(api_config: Optional[Configuration] = None) -> ApiClient:
    """Returns an ApiClient instance."""

    if api_config is None:
        api_config = get_api_config()

    client = ApiClient(api_config)

    client.user_agent = inductiva.get_api_agent()

    return client


def submit_request(api_instance: TasksApi,
                   request: TaskRequest) -> TaskSubmittedInfo:
    """Submits a task request to the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
    Return:
        Returns the body of the HTTP response.
        Contains two fields, "id" and "status".
    """

    api_response = api_instance.submit_task(body=request)

    logging.debug("Request status: %s", api_response.body["status"])

    return api_response.body


def prepare_input(task_id, original_params, type_annotations):
    sim_dir = original_params["sim_dir"]
    # If the input directory is empty, do not zip it
    # still need to zip the input parameters though
    if sim_dir:
        inputs_size = files.get_path_size(sim_dir)
        logging.info("Preparing upload of the local input directory %s (%s).",
                     sim_dir, format_utils.bytes_formatter(inputs_size))

        if os.path.isfile(os.path.join(sim_dir, constants.TASK_OUTPUT_ZIP)):
            raise ValueError(
                f"Invalid file name: '{constants.TASK_OUTPUT_ZIP}'")

    input_zip_path = pack_input(
        params=original_params,
        type_annotations=type_annotations,
        zip_name=task_id,
    )

    zip_file_size = os.path.getsize(input_zip_path)
    logging.info("Input archive size: %s",
                 format_utils.bytes_formatter(zip_file_size))

    logging.info("Uploading input archive...")

    return input_zip_path, zip_file_size


def get_upload_url(
    api_endpoint,
    query_params: Dict[str, str] = None,
    path_params: Dict[str, str] = None,
):
    """
    Fetches the upload URL using the specified method.
    """
    params = {}
    if query_params is not None:
        params["query_params"] = query_params
    if path_params is not None:
        params["path_params"] = path_params

    return api_endpoint(**params).body


def upload_file(api_instance: ApiClient, input_path: str, method: str, url: str,
                progress_bar: tqdm):
    """
    Handles the upload of a file, updating the provided progress bar.
    """
    headers = {"Content-Type": "application/octet-stream"}

    with open(input_path, "rb") as zip_fp:
        wrapped_file = tqdm.utils.CallbackIOWrapper(progress_bar.update, zip_fp,
                                                    "read")
        pool_manager: urllib3.PoolManager = (
            api_instance.api_client.rest_client.pool_manager)
        resp = pool_manager.request(method,
                                    url,
                                    body=wrapped_file,
                                    headers=headers)
        if resp.status != 200:
            raise ApiException(status=resp.status, reason=resp.reason)


def notify_upload_complete(api_endpoint,
                           query_params: Dict[str, str] = None,
                           path_params: Dict[str, str] = None):
    """
    Notifies the API that the file upload is complete.
    """
    params = {}
    if query_params is not None:
        params["query_params"] = query_params
    if path_params is not None:
        params["path_params"] = path_params

    api_endpoint(**params)


def upload_input(api_instance: TasksApi, task_id, original_params,
                 type_annotations):
    """Uploads the inputs of a given task to the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task.
        original_params: Params of the request passed by the user.
        type_annotations: Annotations of the params' types.
    """

    input_zip_path, zip_file_size = prepare_input(task_id, original_params,
                                                  type_annotations)

    api_response = get_upload_url(
        api_instance.get_input_upload_url,
        path_params={"task_id": task_id},
    )

    with tqdm.tqdm(total=zip_file_size,
                   unit="B",
                   unit_scale=True,
                   unit_divisor=1000) as progress_bar:

        method = api_response["method"]
        url = api_response["url"]

        upload_file(api_instance, input_zip_path, method, url, progress_bar)

        notify_upload_complete(
            api_instance.notify_input_uploaded,
            path_params={"task_id": task_id},
        )

    logging.info("Local input directory successfully uploaded.")
    logging.info("")
    os.remove(input_zip_path)


def block_until_finish(api_instance: TasksApi, task_id: str) -> str:
    """Block until a task executing remotely finishes execution.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task to wait for.

    Returns:
        Returns info related to the task, containing two fields,
        "id" and "status".
    """
    logging.debug("Blocking until task is finished ...")
    return block_until_status_is(api_instance, task_id, {"success", "failed"})


def kill_task(api_instance: TasksApi, task_id: str):
    """Kill a task that is executing remotely.

    The function sends a kill request to the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task to kill.
   """
    logging.debug("Sending kill task request ...")
    api_instance.kill_task(path_params={"task_id": task_id},)
    logging.info("Task terminated.")


def get_task_status(api_instance: TasksApi, task_id: str) -> TaskStatus:
    """Check the status of a task."""

    api_response = api_instance.get_task_status(
        path_params={"task_id": task_id})

    status = api_response.body["status"]

    return status


def block_until_status_is(api_instance: TasksApi,
                          task_id,
                          desired_status,
                          sleep_secs=0.5):
    """Block until the status of a task becomes the desired status.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task to wait for.
        desired_status: Task status to wait for.
        sleep_secs: Polling interval.

    Returns:
        Returns info related to the task, containing two fields,
    """
    prev_status = None

    while True:
        try:
            status = get_task_status(api_instance, task_id)
            logging.debug("Task status is %s", status)
        except ApiException as e:
            raise e

        if status != prev_status:
            if status == "submitted":
                logging.info("Waiting for resources...")

            prev_status = status

        # If status reaches the desired status, then stop polling
        if status in desired_status:
            break

        time.sleep(sleep_secs)

    return status


@contextmanager
def blocking_task_context(api_instance: TasksApi, task_id):
    """Context to handle execution of a blocking task.

    The context handles exceptions and the SIGINT signal, issuing a request
    to the API to kill the executing task.
    Info on the implementation:
    - https://docs.python.org/3/library/contextlib.html

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task being executed.
    """
    # Other imported modules can make changes to the SIGINT handler, so we set
    # the default int handler to make sure that KeyboardInterrupt is raised
    # if the user presses Ctrl+C.
    original_sig = signal.signal(signal.SIGINT, signal.default_int_handler)

    try:
        yield None
    except Exception as err:
        logging.info("Caught exception: terminating blocking task...")
        kill_task(api_instance, task_id)
        raise err
    except KeyboardInterrupt as err:
        logging.info("Caught SIGINT: terminating blocking task...")
        kill_task(api_instance, task_id)
        raise err
    finally:
        # Reset original SIGINT handler
        signal.signal(signal.SIGINT, original_sig)


def task_info_str(
    task_id,
    params,
    resource_pool,
    simulator,
    task_submitted_info: TaskSubmittedInfo,
) -> str:
    """Generate a string with the main components of a task submission."""

    info_str = ("■ Task Information:\n"
                f"\t· ID:                    {task_id}\n")
    if simulator is not None:
        info_str += (f"\t· Simulator:             {simulator.name}\n"
                     f"\t· Version:               {simulator.version}\n"
                     f"\t· Image:                 {simulator.image_uri}\n")

    local_input_dir = params["sim_dir"]
    info_str += (f"\t· Local input directory: {local_input_dir}\n"
                 "\t· Submitting to the following computational resources:\n")
    if resource_pool is not None:
        info_str += f" \t\t· {resource_pool}\n"
    else:
        machine_type = constants.DEFAULT_QUEUE_MACHINE_TYPE
        info_str += f" \t\t· Default queue with {machine_type} machines.\n"
        ttl_seconds = task_submitted_info.get("time_to_live_seconds")
        if ttl_seconds is not None and isinstance(ttl_seconds, decimal.Decimal):
            ttl_seconds = format_utils.seconds_formatter(ttl_seconds)
            info_str += (f" \t\t· Task will be killed after the computation "
                         f"time exceeds {ttl_seconds} (h:m:s).\n")
    info_str += "\n"
    return info_str


def submit_task(api_instance,
                simulator,
                request_params,
                resource_pool,
                storage_path_prefix,
                params,
                type_annotations,
                resubmit_on_preemption: bool = False,
                container_image: Optional[str] = None,
                simulator_name_alias: Optional[str] = None,
                simulator_obj=None,
                input_resources: Optional[List[str]] = None):
    """Submit a task and send input files to the API."""
    resource_pool_id = resource_pool.id

    current_project = inductiva.projects.get_current_project()
    if current_project is not None:
        if not current_project.opened:
            raise RuntimeError("Trying to submit a task to a closed project.")
        current_project = current_project.name

    if not input_resources:
        input_resources = []
    task_request = TaskRequest(simulator=simulator,
                               params=request_params,
                               project=current_project,
                               resource_pool=resource_pool_id,
                               container_image=container_image,
                               storage_path_prefix=storage_path_prefix,
                               simulator_name_alias=simulator_name_alias,
                               resubmit_on_preemption=resubmit_on_preemption,
                               input_resources=input_resources)

    task_submitted_info = submit_request(
        api_instance=api_instance,
        request=task_request,
    )

    task_id = task_submitted_info["id"]
    logging.info(
        task_info_str(
            task_id,
            params,
            resource_pool,
            simulator_obj,
            task_submitted_info,
        ))

    if task_submitted_info["status"] == "pending-input":

        upload_input(
            api_instance=api_instance,
            original_params=params,
            task_id=task_id,
            type_annotations=type_annotations,
        )

    return task_id


def invoke_async_api(simulator: str,
                     params,
                     type_annotations: Dict[Any, Type],
                     resource_pool: types.ComputationalResources,
                     storage_path_prefix: Optional[str] = "",
                     container_image: Optional[str] = None,
                     resubmit_on_preemption: bool = False,
                     simulator_obj=None,
                     simulator_name_alias=None,
                     input_resources: Optional[List[str]] = None) -> str:
    """Perform a task asyc and remotely via Inductiva's Web API.

    Submits a simulation async to the API and returns the task id.
    The flow is summarized as follows:
        1. Transform request params into the params used to
        validate permission to execute the request.
        2. Submit task via the "POST task/submit" endpoint.
            Note: HTTP status code 400 informs the requested method is invalid,
                and HTTP status code 403 informs that the user is not authorized
                to post such request.
        3. If the status returned by the previous HTTP request is
            "pending-input", ZIP inputs and send them via
            "POST task/{task_id}/input".
        4. Return task_id and leaves the simulation on the queue until resources
            become available.

    Args:
        request: Request sent to the API for validation.
        input_dir: Directory containing the input files to be uploaded.
        container_image: The container image to use for the simulation
            Example: container_image="docker://inductiva/kutu:xbeach_v1.23_dev"
        resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
        input_resources: Additional input files that will be copied to the
                simulation from a bucket or from another task output.
    Return:
        Returns the task id.
    """

    request_params = get_validate_request_params(
        original_params=params,
        type_annotations=type_annotations,
    )

    with get_client() as client:
        api_instance = TasksApi(client)

        task_id = submit_task(api_instance=api_instance,
                              simulator=simulator,
                              simulator_name_alias=simulator_name_alias,
                              request_params=request_params,
                              resource_pool=resource_pool,
                              storage_path_prefix=storage_path_prefix,
                              params=params,
                              container_image=container_image,
                              type_annotations=type_annotations,
                              resubmit_on_preemption=resubmit_on_preemption,
                              simulator_obj=simulator_obj,
                              input_resources=input_resources)

    return task_id



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class TierQuotaDetail(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Mixin to handle decimal.Decimal("Infinity") values. This is useful
when we want to represent infinity in a JSON response. Pydantic does
not support this out of the box. This mixin allows us to use
decimal.Decimal("Infinity") in our models, and it will be serialized to
"infinity" in the JSON response.
    """

    class MetaOapg:
        required = {
            "unit",
            "scope",
            "id",
            "label",
            "value",
        }

        class properties:
            id = schemas.StrSchema
            value = schemas.NumberSchema
            label = schemas.StrSchema
            unit = schemas.StrSchema
            scope = schemas.StrSchema
            __annotations__ = {
                "id": id,
                "value": value,
                "label": label,
                "unit": unit,
                "scope": scope,
            }

    unit: MetaOapg.properties.unit
    scope: MetaOapg.properties.scope
    id: MetaOapg.properties.id
    label: MetaOapg.properties.label
    value: MetaOapg.properties.value

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["value"]
    ) -> MetaOapg.properties.value:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["label"]
    ) -> MetaOapg.properties.label:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["unit"]
    ) -> MetaOapg.properties.unit:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["scope"]
    ) -> MetaOapg.properties.scope:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "id",
        "value",
        "label",
        "unit",
        "scope",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["value"]
    ) -> MetaOapg.properties.value:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["label"]
    ) -> MetaOapg.properties.label:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["unit"]
    ) -> MetaOapg.properties.unit:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["scope"]
    ) -> MetaOapg.properties.scope:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "id",
        "value",
        "label",
        "unit",
        "scope",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        unit: typing.Union[
            MetaOapg.properties.unit,
            str,
        ],
        scope: typing.Union[
            MetaOapg.properties.scope,
            str,
        ],
        id: typing.Union[
            MetaOapg.properties.id,
            str,
        ],
        label: typing.Union[
            MetaOapg.properties.label,
            str,
        ],
        value: typing.Union[
            MetaOapg.properties.value,
            decimal.Decimal,
            int,
            float,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'TierQuotaDetail':
        return super().__new__(
            cls,
            *_args,
            unit=unit,
            scope=scope,
            id=id,
            label=label,
            value=value,
            _configuration=_configuration,
            **kwargs,
        )



"""Class to run commands on an custom image."""
from typing import List, Optional

from inductiva import types, tasks, simulators


@simulators.simulator.mpi_enabled
class CustomImage(simulators.Simulator):
    """Class to run commands on an custom image."""

    def __init__(self, container_image: str):
        """Initialize the ArbitraryImage class.
        Point to the API method to run a simulation.
        Args:
            container_image: The container image to use for the simulation.
                Example: container_image="docker://inductiva/kutu:xbeach_v1.23"
        """
        self.container_image = container_image
        super().__init__()

        self.simulator = "arbitrary_commands"

    def _get_image_uri(self):
        return self.container_image

    def run(self,
            input_dir: Optional[str],
            commands: List[str],
            *,
            on: types.ComputationalResources,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[List[str]] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.
        Args:
            input_dir: Path to the directory containing the input files.
            commands: List of commands to run.
            on: The computational resource to launch the simulation on.
            storage_dir: Parent directory for storing simulation
                               results.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
        """
        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           container_image=self._image_uri,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           **kwargs)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class TaskRunnerRegisterInfo(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Info for creating a task-runner.
    """

    class MetaOapg:
        required = {
            "memory",
            "create_time",
            "cpu_count_logical",
            "cpu_count_physical",
            "host_id",
            "host_name",
            "machine_group_id",
        }

        class properties:
            machine_group_id = schemas.UUIDSchema
            create_time = schemas.DateTimeSchema
            cpu_count_logical = schemas.IntSchema
            cpu_count_physical = schemas.IntSchema
            memory = schemas.IntSchema
            host_name = schemas.StrSchema
            host_id = schemas.StrSchema
            mpi_cluster = schemas.BoolSchema
            __annotations__ = {
                "machine_group_id": machine_group_id,
                "create_time": create_time,
                "cpu_count_logical": cpu_count_logical,
                "cpu_count_physical": cpu_count_physical,
                "memory": memory,
                "host_name": host_name,
                "host_id": host_id,
                "mpi_cluster": mpi_cluster,
            }

    memory: MetaOapg.properties.memory
    create_time: MetaOapg.properties.create_time
    cpu_count_logical: MetaOapg.properties.cpu_count_logical
    cpu_count_physical: MetaOapg.properties.cpu_count_physical
    host_id: MetaOapg.properties.host_id
    host_name: MetaOapg.properties.host_name
    machine_group_id: MetaOapg.properties.machine_group_id

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["machine_group_id"]
    ) -> MetaOapg.properties.machine_group_id:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["create_time"]
    ) -> MetaOapg.properties.create_time:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["cpu_count_logical"]
    ) -> MetaOapg.properties.cpu_count_logical:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["cpu_count_physical"]
    ) -> MetaOapg.properties.cpu_count_physical:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["memory"]
    ) -> MetaOapg.properties.memory:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["host_name"]
    ) -> MetaOapg.properties.host_name:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["host_id"]
    ) -> MetaOapg.properties.host_id:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["mpi_cluster"]
    ) -> MetaOapg.properties.mpi_cluster:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "machine_group_id",
        "create_time",
        "cpu_count_logical",
        "cpu_count_physical",
        "memory",
        "host_name",
        "host_id",
        "mpi_cluster",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["machine_group_id"]
    ) -> MetaOapg.properties.machine_group_id:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["create_time"]
    ) -> MetaOapg.properties.create_time:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["cpu_count_logical"]
    ) -> MetaOapg.properties.cpu_count_logical:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["cpu_count_physical"]
    ) -> MetaOapg.properties.cpu_count_physical:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["memory"]
    ) -> MetaOapg.properties.memory:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["host_name"]
    ) -> MetaOapg.properties.host_name:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["host_id"]
    ) -> MetaOapg.properties.host_id:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["mpi_cluster"]
    ) -> typing.Union[MetaOapg.properties.mpi_cluster, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "machine_group_id",
        "create_time",
        "cpu_count_logical",
        "cpu_count_physical",
        "memory",
        "host_name",
        "host_id",
        "mpi_cluster",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        memory: typing.Union[
            MetaOapg.properties.memory,
            decimal.Decimal,
            int,
        ],
        create_time: typing.Union[
            MetaOapg.properties.create_time,
            str,
            datetime,
        ],
        cpu_count_logical: typing.Union[
            MetaOapg.properties.cpu_count_logical,
            decimal.Decimal,
            int,
        ],
        cpu_count_physical: typing.Union[
            MetaOapg.properties.cpu_count_physical,
            decimal.Decimal,
            int,
        ],
        host_id: typing.Union[
            MetaOapg.properties.host_id,
            str,
        ],
        host_name: typing.Union[
            MetaOapg.properties.host_name,
            str,
        ],
        machine_group_id: typing.Union[
            MetaOapg.properties.machine_group_id,
            str,
            uuid.UUID,
        ],
        mpi_cluster: typing.Union[MetaOapg.properties.mpi_cluster, bool,
                                  schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'TaskRunnerRegisterInfo':
        return super().__new__(
            cls,
            *_args,
            memory=memory,
            create_time=create_time,
            cpu_count_logical=cpu_count_logical,
            cpu_count_physical=cpu_count_physical,
            host_id=host_id,
            host_name=host_name,
            machine_group_id=machine_group_id,
            mpi_cluster=mpi_cluster,
            _configuration=_configuration,
            **kwargs,
        )



"""Utility to run a function at regular time intervals."""
from typing import Union, Callable
import threading
import time


class StoppableScheduler(threading.Thread):
    """A stoppable scheduler for running a function at regular intervals.
    
    The scheduler runs a given function at regular intervals until the stop
    method is called. If the call to the action takes longer than the
    interval, the scheduler will wait for the remaining time before calling
    the action at the next interval.
    """

    def __init__(self,
                 every: Union[int, float],
                 action: Callable[..., None],
                 args=(),
                 kwargs=None):
        """Initialize the scheduler.

        Args:
            every (float): The interval at which the action should be executed.
            action (callable): The function to be executed.
            args (tuple, optional): The positional arguments to be passed to the
                action. Defaults to ().
            kwargs (dict, optional): The keyword arguments to be passed to the
                action. Defaults to None.
        """
        super().__init__()
        self.every = every
        self.action = action
        self.args = args
        self.kwargs = kwargs or {}

        self._event = threading.Event()
        self._count = 0

    @property
    def count(self) -> int:
        """Return the number of times the action has been executed."""
        return self._count

    def run(self) -> None:
        """Run the scheduler until the stop method is called."""
        every = self.every

        while not self._event.is_set():
            now = time.time()
            self._count += 1
            self.action(*self.args, **self.kwargs)
            elapsed = time.time() - now

            # take into account the time the action takes to complete
            #if elapsed < self.every:
            #    self._event.wait(self.every - elapsed)
            self._event.wait(every - elapsed % every)

    def stop(self) -> None:
        """Stop the scheduler."""
        self._event.set()



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class MachineType(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "machine_type",
            "num_cpus",
            "price",
            "ram_gb",
        }

        class properties:
            machine_type = schemas.StrSchema
            num_cpus = schemas.IntSchema
            ram_gb = schemas.IntSchema
            price = schemas.NumberSchema

            class provider_id(
                    schemas.ComposedSchema,):

                class MetaOapg:

                    @classmethod
                    @functools.lru_cache()
                    def all_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            Providers,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'provider_id':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class threads_per_core(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'threads_per_core':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class spot(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.BoolSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'spot':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class region(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'region':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class zone(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'zone':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "machine_type": machine_type,
                "num_cpus": num_cpus,
                "ram_gb": ram_gb,
                "price": price,
                "provider_id": provider_id,
                "threads_per_core": threads_per_core,
                "spot": spot,
                "region": region,
                "zone": zone,
            }

    machine_type: MetaOapg.properties.machine_type
    num_cpus: MetaOapg.properties.num_cpus
    price: MetaOapg.properties.price
    ram_gb: MetaOapg.properties.ram_gb

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["num_cpus"]
    ) -> MetaOapg.properties.num_cpus:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["ram_gb"]
    ) -> MetaOapg.properties.ram_gb:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["price"]
    ) -> MetaOapg.properties.price:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["provider_id"]
    ) -> MetaOapg.properties.provider_id:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["threads_per_core"]
    ) -> MetaOapg.properties.threads_per_core:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["spot"]
    ) -> MetaOapg.properties.spot:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["region"]
    ) -> MetaOapg.properties.region:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["zone"]
    ) -> MetaOapg.properties.zone:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "machine_type",
        "num_cpus",
        "ram_gb",
        "price",
        "provider_id",
        "threads_per_core",
        "spot",
        "region",
        "zone",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["num_cpus"]
    ) -> MetaOapg.properties.num_cpus:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["ram_gb"]
    ) -> MetaOapg.properties.ram_gb:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["price"]
    ) -> MetaOapg.properties.price:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["provider_id"]
    ) -> typing.Union[MetaOapg.properties.provider_id, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["threads_per_core"]
    ) -> typing.Union[MetaOapg.properties.threads_per_core, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["spot"]
    ) -> typing.Union[MetaOapg.properties.spot, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["region"]
    ) -> typing.Union[MetaOapg.properties.region, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["zone"]
    ) -> typing.Union[MetaOapg.properties.zone, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "machine_type",
        "num_cpus",
        "ram_gb",
        "price",
        "provider_id",
        "threads_per_core",
        "spot",
        "region",
        "zone",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        machine_type: typing.Union[
            MetaOapg.properties.machine_type,
            str,
        ],
        num_cpus: typing.Union[
            MetaOapg.properties.num_cpus,
            decimal.Decimal,
            int,
        ],
        price: typing.Union[
            MetaOapg.properties.price,
            decimal.Decimal,
            int,
            float,
        ],
        ram_gb: typing.Union[
            MetaOapg.properties.ram_gb,
            decimal.Decimal,
            int,
        ],
        provider_id: typing.Union[MetaOapg.properties.provider_id, dict,
                                  frozendict.frozendict, str, date, datetime,
                                  uuid.UUID, int, float, decimal.Decimal, bool,
                                  None, list, tuple, bytes, io.FileIO,
                                  io.BufferedReader,
                                  schemas.Unset] = schemas.unset,
        threads_per_core: typing.Union[MetaOapg.properties.threads_per_core,
                                       dict, frozendict.frozendict, str, date,
                                       datetime, uuid.UUID, int, float,
                                       decimal.Decimal, bool, None, list, tuple,
                                       bytes, io.FileIO, io.BufferedReader,
                                       schemas.Unset] = schemas.unset,
        spot: typing.Union[MetaOapg.properties.spot, dict,
                           frozendict.frozendict, str, date, datetime,
                           uuid.UUID, int, float, decimal.Decimal, bool, None,
                           list, tuple, bytes, io.FileIO, io.BufferedReader,
                           schemas.Unset] = schemas.unset,
        region: typing.Union[MetaOapg.properties.region, dict,
                             frozendict.frozendict, str, date, datetime,
                             uuid.UUID, int, float, decimal.Decimal, bool, None,
                             list, tuple, bytes, io.FileIO, io.BufferedReader,
                             schemas.Unset] = schemas.unset,
        zone: typing.Union[MetaOapg.properties.zone, dict,
                           frozendict.frozendict, str, date, datetime,
                           uuid.UUID, int, float, decimal.Decimal, bool, None,
                           list, tuple, bytes, io.FileIO, io.BufferedReader,
                           schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'MachineType':
        return super().__new__(
            cls,
            *_args,
            machine_type=machine_type,
            num_cpus=num_cpus,
            price=price,
            ram_gb=ram_gb,
            provider_id=provider_id,
            threads_per_core=threads_per_core,
            spot=spot,
            region=region,
            zone=zone,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.providers import Providers



"""OpenFOAM module of the API for fluid dynamics."""
from typing import List, Optional

from inductiva import types, tasks, simulators

AVAILABLE_OPENFOAM_DISTRIBUTIONS = ["foundation", "esi"]


@simulators.simulator.mpi_enabled
class OpenFOAM(simulators.Simulator):
    """Class to invoke a generic OpenFOAM simulation on the API.

    Users can choose between the ESI or the Foundation version
    by selecting the version on the initiliasation. Be aware, that
    some input files may only work for a specific version.
    """

    def __init__(self,
                 /,
                 distribution: str = "foundation",
                 version: Optional[str] = None,
                 use_dev: bool = False):
        """Initialize the OpenFOAM simulator.

        Args:
            distribution (str): The distribution of OpenFOAM to use. Available
                distributions are "foundation" and "esi". Default is
                "foundation".
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        if distribution not in AVAILABLE_OPENFOAM_DISTRIBUTIONS:
            raise ValueError(
                f"Distribution '{distribution}' of OpenFOAM is not supported. "
                f"Available distributions are: "
                f"{AVAILABLE_OPENFOAM_DISTRIBUTIONS}")

        self._distribution = distribution

        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = f"openfoam_{distribution}"
        self.container_image = self._get_image_uri()

    @property
    def name(self):
        """Get the name of the simulator."""
        return "OpenFOAM-" + self._distribution

    def run(self,
            input_dir: Optional[str],
            *,
            commands: Optional[List[types.Commands]] = None,
            shell_script: Optional[str] = None,
            on: types.ComputationalResources,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[List[str]] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            commands: List of commands to run using the OpenFOAM simulator.
            shell_script: Path to a shell script (relative to input_dir) to run
                the simulation.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            other arguments: See the documentation of the base class.
        """
        if not commands and not shell_script:
            raise ValueError("Either 'commands' or 'shell_script'"
                             " must be provided.")
        if commands and shell_script:
            raise ValueError("Only one of 'commands' or 'shell_script'"
                             " must be provided.")

        if shell_script:
            commands = [f"bash {shell_script}"]

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           **kwargs)



"""Main CLI entrypoint."""
from typing import TextIO
import argparse
import sys
import os
import re
import io

import inductiva
from inductiva import _cli
from inductiva import constants, utils
from . import loader
try:
    from . import ansi_pager
except ImportError:
    ansi_pager = None


def get_main_parser():
    parser = argparse.ArgumentParser(
        prog="inductiva",
        description="CLI tool for Inductiva API.",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {inductiva.__version__}",
    )

    # If no subcommand is provided, print help
    _cli.utils.show_help_msg(parser)

    # Create subcommands
    subparsers = parser.add_subparsers(title="available subcomands")

    # Load all modules starting with "cmd_" as subcommands.
    loader.load_commands(subparsers,
                         os.path.dirname(__file__),
                         "inductiva._cli",
                         prefix=constants.LOADER_COMMAND_PREFIX)
    return parser


def watch(func, every, args, cmd):
    """Run the function at regular intervals and display results in a pager."""
    cmd = re.sub(r"(-w|--watch)\s*([+-]?([0-9]*[.])?[0-9]+)?\s*", "", cmd, 1)
    header = f"> every {every}s: inductiva {cmd}"

    def action(fout: TextIO = sys.stdout):
        buffer = io.StringIO()
        func(args, fout=buffer)
        fout.clear()
        fout.write(buffer.getvalue())

    with ansi_pager.PagedOutput(header) as pager:
        scheduler = utils.scheduler.StoppableScheduler(every,
                                                       action,
                                                       args=(pager,))
        scheduler.daemon = True
        scheduler.start()
        pager.run()
        scheduler.stop()
        scheduler.join(0.5)


def main():
    parser = get_main_parser()
    args = parser.parse_args()

    exit_code = 0
    # Call the function associated with the subcommand
    if getattr(args, "watchable", False) and args.watch is not None:
        if ansi_pager is None:
            raise ImportError("watch module is not available.")
        else:
            cmd = " ".join(sys.argv[1:])
            watch(args.func, args.watch, args, cmd)
    else:
        exit_code = args.func(args)
    return exit_code



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class MachineGroupWithCostsAdminView(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "machine_type",
            "costs",
            "num_vms",
            "creation_timestamp",
            "type",
            "disk_size_gb",
            "threads_per_core",
            "spot",
            "name",
            "provider_id",
            "id",
            "user",
            "statistics",
            "status",
        }

        class properties:
            id = schemas.UUIDSchema
            name = schemas.StrSchema
            machine_type = schemas.StrSchema
            disk_size_gb = schemas.IntSchema
            num_vms = schemas.IntSchema
            spot = schemas.BoolSchema

            @staticmethod
            def type() -> typing.Type['MachineGroupType']:
                return MachineGroupType

            creation_timestamp = schemas.DateTimeSchema

            @staticmethod
            def provider_id() -> typing.Type['Providers']:
                return Providers

            @staticmethod
            def status() -> typing.Type['MachineGroupStatus']:
                return MachineGroupStatus

            threads_per_core = schemas.IntSchema

            @staticmethod
            def statistics() -> typing.Type['MachineGroupUsageStatistics']:
                return MachineGroupUsageStatistics

            @staticmethod
            def costs() -> typing.Type['MachineGroupCostsAdminView']:
                return MachineGroupCostsAdminView

            @staticmethod
            def user() -> typing.Type['UserEmailOrganization']:
                return UserEmailOrganization

            class auto_resize_disk_max_gb(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'auto_resize_disk_max_gb':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class min_vms(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'min_vms':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class max_vms(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'max_vms':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class start_timestamp(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'start_timestamp':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class deletion_timestamp(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'deletion_timestamp':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class provider_description(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'provider_description':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class max_idle_time(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    format = 'time-delta'
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'max_idle_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class termination_reason(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'termination_reason':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "id": id,
                "name": name,
                "machine_type": machine_type,
                "disk_size_gb": disk_size_gb,
                "num_vms": num_vms,
                "spot": spot,
                "type": type,
                "creation_timestamp": creation_timestamp,
                "provider_id": provider_id,
                "status": status,
                "threads_per_core": threads_per_core,
                "statistics": statistics,
                "costs": costs,
                "user": user,
                "auto_resize_disk_max_gb": auto_resize_disk_max_gb,
                "min_vms": min_vms,
                "max_vms": max_vms,
                "start_timestamp": start_timestamp,
                "deletion_timestamp": deletion_timestamp,
                "provider_description": provider_description,
                "max_idle_time": max_idle_time,
                "termination_reason": termination_reason,
            }

    machine_type: MetaOapg.properties.machine_type
    costs: 'MachineGroupCostsAdminView'
    num_vms: MetaOapg.properties.num_vms
    creation_timestamp: MetaOapg.properties.creation_timestamp
    type: 'MachineGroupType'
    disk_size_gb: MetaOapg.properties.disk_size_gb
    threads_per_core: MetaOapg.properties.threads_per_core
    spot: MetaOapg.properties.spot
    name: MetaOapg.properties.name
    provider_id: 'Providers'
    id: MetaOapg.properties.id
    user: 'UserEmailOrganization'
    statistics: 'MachineGroupUsageStatistics'
    status: 'MachineGroupStatus'

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["disk_size_gb"]
    ) -> MetaOapg.properties.disk_size_gb:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["num_vms"]
    ) -> MetaOapg.properties.num_vms:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["spot"]
    ) -> MetaOapg.properties.spot:
        ...

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["type"]) -> 'MachineGroupType':
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["creation_timestamp"]
    ) -> MetaOapg.properties.creation_timestamp:
        ...

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["provider_id"]) -> 'Providers':
        ...

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["status"]) -> 'MachineGroupStatus':
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["threads_per_core"]
    ) -> MetaOapg.properties.threads_per_core:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["statistics"]
    ) -> 'MachineGroupUsageStatistics':
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["costs"]
    ) -> 'MachineGroupCostsAdminView':
        ...

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["user"]) -> 'UserEmailOrganization':
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["auto_resize_disk_max_gb"]
    ) -> MetaOapg.properties.auto_resize_disk_max_gb:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["min_vms"]
    ) -> MetaOapg.properties.min_vms:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["max_vms"]
    ) -> MetaOapg.properties.max_vms:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["start_timestamp"]
    ) -> MetaOapg.properties.start_timestamp:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["deletion_timestamp"]
    ) -> MetaOapg.properties.deletion_timestamp:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["provider_description"]
    ) -> MetaOapg.properties.provider_description:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["max_idle_time"]
    ) -> MetaOapg.properties.max_idle_time:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["termination_reason"]
    ) -> MetaOapg.properties.termination_reason:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "id",
        "name",
        "machine_type",
        "disk_size_gb",
        "num_vms",
        "spot",
        "type",
        "creation_timestamp",
        "provider_id",
        "status",
        "threads_per_core",
        "statistics",
        "costs",
        "user",
        "auto_resize_disk_max_gb",
        "min_vms",
        "max_vms",
        "start_timestamp",
        "deletion_timestamp",
        "provider_description",
        "max_idle_time",
        "termination_reason",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["disk_size_gb"]
    ) -> MetaOapg.properties.disk_size_gb:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["num_vms"]
    ) -> MetaOapg.properties.num_vms:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["spot"]
    ) -> MetaOapg.properties.spot:
        ...

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["type"]) -> 'MachineGroupType':
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["creation_timestamp"]
    ) -> MetaOapg.properties.creation_timestamp:
        ...

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["provider_id"]) -> 'Providers':
        ...

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["status"]) -> 'MachineGroupStatus':
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["threads_per_core"]
    ) -> MetaOapg.properties.threads_per_core:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["statistics"]
    ) -> 'MachineGroupUsageStatistics':
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["costs"]
    ) -> 'MachineGroupCostsAdminView':
        ...

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["user"]) -> 'UserEmailOrganization':
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["auto_resize_disk_max_gb"]
    ) -> typing.Union[MetaOapg.properties.auto_resize_disk_max_gb,
                      schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["min_vms"]
    ) -> typing.Union[MetaOapg.properties.min_vms, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["max_vms"]
    ) -> typing.Union[MetaOapg.properties.max_vms, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["start_timestamp"]
    ) -> typing.Union[MetaOapg.properties.start_timestamp, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["deletion_timestamp"]
    ) -> typing.Union[MetaOapg.properties.deletion_timestamp, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["provider_description"]
    ) -> typing.Union[MetaOapg.properties.provider_description, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["max_idle_time"]
    ) -> typing.Union[MetaOapg.properties.max_idle_time, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["termination_reason"]
    ) -> typing.Union[MetaOapg.properties.termination_reason, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "id",
        "name",
        "machine_type",
        "disk_size_gb",
        "num_vms",
        "spot",
        "type",
        "creation_timestamp",
        "provider_id",
        "status",
        "threads_per_core",
        "statistics",
        "costs",
        "user",
        "auto_resize_disk_max_gb",
        "min_vms",
        "max_vms",
        "start_timestamp",
        "deletion_timestamp",
        "provider_description",
        "max_idle_time",
        "termination_reason",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        machine_type: typing.Union[
            MetaOapg.properties.machine_type,
            str,
        ],
        costs: 'MachineGroupCostsAdminView',
        num_vms: typing.Union[
            MetaOapg.properties.num_vms,
            decimal.Decimal,
            int,
        ],
        creation_timestamp: typing.Union[
            MetaOapg.properties.creation_timestamp,
            str,
            datetime,
        ],
        type: 'MachineGroupType',
        disk_size_gb: typing.Union[
            MetaOapg.properties.disk_size_gb,
            decimal.Decimal,
            int,
        ],
        threads_per_core: typing.Union[
            MetaOapg.properties.threads_per_core,
            decimal.Decimal,
            int,
        ],
        spot: typing.Union[
            MetaOapg.properties.spot,
            bool,
        ],
        name: typing.Union[
            MetaOapg.properties.name,
            str,
        ],
        provider_id: 'Providers',
        id: typing.Union[
            MetaOapg.properties.id,
            str,
            uuid.UUID,
        ],
        user: 'UserEmailOrganization',
        statistics: 'MachineGroupUsageStatistics',
        status: 'MachineGroupStatus',
        auto_resize_disk_max_gb: typing.Union[
            MetaOapg.properties.auto_resize_disk_max_gb, dict,
            frozendict.frozendict, str, date, datetime, uuid.UUID, int, float,
            decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO,
            io.BufferedReader, schemas.Unset] = schemas.unset,
        min_vms: typing.Union[MetaOapg.properties.min_vms, dict,
                              frozendict.frozendict, str, date, datetime,
                              uuid.UUID, int, float, decimal.Decimal, bool,
                              None, list, tuple, bytes, io.FileIO,
                              io.BufferedReader, schemas.Unset] = schemas.unset,
        max_vms: typing.Union[MetaOapg.properties.max_vms, dict,
                              frozendict.frozendict, str, date, datetime,
                              uuid.UUID, int, float, decimal.Decimal, bool,
                              None, list, tuple, bytes, io.FileIO,
                              io.BufferedReader, schemas.Unset] = schemas.unset,
        start_timestamp: typing.Union[MetaOapg.properties.start_timestamp, dict,
                                      frozendict.frozendict, str, date,
                                      datetime, uuid.UUID, int, float,
                                      decimal.Decimal, bool, None, list, tuple,
                                      bytes, io.FileIO, io.BufferedReader,
                                      schemas.Unset] = schemas.unset,
        deletion_timestamp: typing.Union[MetaOapg.properties.deletion_timestamp,
                                         dict, frozendict.frozendict, str, date,
                                         datetime, uuid.UUID, int, float,
                                         decimal.Decimal, bool, None, list,
                                         tuple, bytes, io.FileIO,
                                         io.BufferedReader,
                                         schemas.Unset] = schemas.unset,
        provider_description: typing.Union[
            MetaOapg.properties.provider_description, dict,
            frozendict.frozendict, str, date, datetime, uuid.UUID, int, float,
            decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO,
            io.BufferedReader, schemas.Unset] = schemas.unset,
        max_idle_time: typing.Union[MetaOapg.properties.max_idle_time, dict,
                                    frozendict.frozendict, str, date, datetime,
                                    uuid.UUID, int, float, decimal.Decimal,
                                    bool, None, list, tuple, bytes, io.FileIO,
                                    io.BufferedReader,
                                    schemas.Unset] = schemas.unset,
        termination_reason: typing.Union[MetaOapg.properties.termination_reason,
                                         dict, frozendict.frozendict, str, date,
                                         datetime, uuid.UUID, int, float,
                                         decimal.Decimal, bool, None, list,
                                         tuple, bytes, io.FileIO,
                                         io.BufferedReader,
                                         schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'MachineGroupWithCostsAdminView':
        return super().__new__(
            cls,
            *_args,
            machine_type=machine_type,
            costs=costs,
            num_vms=num_vms,
            creation_timestamp=creation_timestamp,
            type=type,
            disk_size_gb=disk_size_gb,
            threads_per_core=threads_per_core,
            spot=spot,
            name=name,
            provider_id=provider_id,
            id=id,
            user=user,
            statistics=statistics,
            status=status,
            auto_resize_disk_max_gb=auto_resize_disk_max_gb,
            min_vms=min_vms,
            max_vms=max_vms,
            start_timestamp=start_timestamp,
            deletion_timestamp=deletion_timestamp,
            provider_description=provider_description,
            max_idle_time=max_idle_time,
            termination_reason=termination_reason,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.machine_group_costs_admin_view import MachineGroupCostsAdminView
from inductiva.client.model.machine_group_status import MachineGroupStatus
from inductiva.client.model.machine_group_type import MachineGroupType
from inductiva.client.model.machine_group_usage_statistics import MachineGroupUsageStatistics
from inductiva.client.model.providers import Providers
from inductiva.client.model.user_email_organization import UserEmailOrganization



"""API for Benchmarking"""
import enum
import json
import csv
import logging
import concurrent.futures
import contextvars
from typing import Optional, Union
from typing_extensions import Self
from collections import defaultdict
from inductiva import types, resources, projects, simulators, client
from inductiva.client.models import TaskStatusCode
from inductiva.utils.format_utils import CURRENCY_SYMBOL, TIME_UNIT


class ExportFormat(enum.Enum):
    """Enumeration of supported benchmark export formats."""
    JSON = "json"
    CSV = "csv"


class SelectMode(enum.Enum):
    """
    Enumeration of supported data selection modes, specifying which data 
    should be included in the benchmarking results.
    """
    ALL = "all"
    DISTINCT = "distinct"


class Benchmark(projects.Project):
    """Represents the benchmark runner."""

    class InfoKey:
        """Constant keys for the info used in the benchmark runs."""
        TASK_ID = "task_id"
        SIMULATOR = "simulator"
        MACHINE_TYPE = "machine_type"
        TIME = f"computation_time ({TIME_UNIT})"
        COST = f"estimated_computation_cost ({CURRENCY_SYMBOL})"

    def __init__(self, name: str, append: bool = True):
        """
        Initializes a new Benchmark instance.

        Args:
            name (str): The name of the benchmark runner.
            append (bool): Indicates whether to allow adding runs to the 
            existing benchmark (default is True).
        """
        super().__init__(name=name, append=append)
        self.runs = []
        self.simulator = None
        self.input_dir = None
        self.on = None
        self.kwargs = {}

    def set_default(
        self,
        simulator: Optional[simulators.Simulator] = None,
        input_dir: Optional[str] = None,
        on: Optional[types.ComputationalResources] = None,
        **kwargs,
    ) -> Self:
        """
        Sets default parameters for the benchmark runner.

        This method allows you to configure default settings for the benchmark,
        which will be used in subsequent runs unless explicitly overridden.

        Args:
            simulator (Optional[Simulator]): The simulator instance to be used
                as the default for future runs. If not provided, the current
                simulator will remain unchanged.
            input_dir (Optional[str]): The directory path for input files. If
                not provided, the current input directory will remain unchanged.
            on (Optional[types.ComputationalResources]): The computational
                resources to use for running the simulations. If not specified,
                the current resources will remain unchanged.
            **kwargs: Additional keyword arguments to set as default parameters 
                for the simulations. These will update any existing parameters
                with the same names.

        Returns:
            Self: The current instance for method chaining.
        """
        self.simulator = simulator or self.simulator
        self.input_dir = input_dir or self.input_dir
        self.on = on or self.on
        self.kwargs = {**self.kwargs, **kwargs}
        return self

    def add_run(
        self,
        simulator: Optional[simulators.Simulator] = None,
        input_dir: Optional[str] = None,
        on: Optional[types.ComputationalResources] = None,
        **kwargs,
    ) -> Self:
        """
        Adds a simulation run to the benchmark.

        Args:
            simulator (Optional[Simulator]): The simulator to be used for this
                run. If not provided, the previously set simulator will be used.
            input_dir (Optional[str]): The directory containing input files for
                the simulation. If not provided, the previously set input
                directory will be used.
            on (Optional[types.ComputationalResources]): The computational
                resources to run the simulation on. If not provided, the
                previously set resources will be used.
            **kwargs: Additional keyword arguments for the simulator run. These
                will overwrite any previously set parameters.

        Returns:
            Self: The current instance for method chaining.
        """
        self.runs.append((
            simulator or self.simulator,
            input_dir or self.input_dir,
            on or self.on,
            {
                **self.kwargs,
                **kwargs
            },
        ))
        return self

    def run(self, num_repeats: int = 2, wait_for_quotas: bool = True) -> Self:
        """
        Executes all added runs.

        Each run is executed the specified number of times, and the collection
        of runs is cleared afterwards.

        Args:
            num_repeats (int): The number of times to repeat each simulation
                run (default is 2).
            wait_for_quotas (bool): Indicates whether to wait for quotas to 
                become available before starting each resource. If `True`, the 
                program will actively wait in a loop, periodically sleeping and 
                checking for quotas. If `False`, the program crashes if quotas 
                are not available (default is `True`).

        Returns:
            Self: The current instance for method chaining.
        """

        def _run(params):
            simulator, input_dir, machine_group, kwargs = params
            if not machine_group.started:
                machine_group.start(wait_for_quotas=wait_for_quotas)
            for _ in range(num_repeats):
                simulator.run(input_dir=input_dir, on=machine_group, **kwargs)

        def _initializer(ctx):
            for var, val in ctx.items():
                var.set(val)

        with self, concurrent.futures.ThreadPoolExecutor(
                initializer=_initializer,
                initargs=(contextvars.copy_context(),)) as executor:
            _ = list(executor.map(_run, self.runs))
        self.runs.clear()
        return self

    def wait(self) -> Self:
        """
        Waits for all running tasks to complete.

        Returns:
            Self: The current instance for method chaining.
        """
        tasks = self.get_tasks()
        for task in tasks:
            task.wait(download_std_on_completion=False)
        return self

    def export(
        self,
        fmt: Union[ExportFormat, str] = ExportFormat.JSON,
        filename: Optional[str] = None,
        status: Optional[Union[TaskStatusCode, str]] = None,
        select: Union[SelectMode, str] = SelectMode.DISTINCT,
    ):
        """
        Exports the benchmark performance metrics in the specified format.

        Args:
            fmt (Union[ExportFormat, str]): The format to export the results
                in. Defaults to ExportFormat.JSON.
            filename (Optional[str]): The name of the output file to save the
                exported results. Defaults to the benchmark's name if not
                provided.
            status (Optional[Union[TaskStatusCode, str]]): The status of the
                tasks to include in the benchmarking results. Defaults to None,
                which includes all tasks.
            select (Union[SelectMode, str]): The data to include in
                the benchmarking results. Defaults to SelectMode.DISTINCT that
                includes only the parameters that vary between different runs.
        """
        if isinstance(fmt, str):
            fmt = ExportFormat[fmt.upper()]
        info = self.runs_info(status=status, select=select)
        filename = filename or f"{self.name}.{fmt.value}"
        if fmt == ExportFormat.JSON:
            with open(filename, mode="w", encoding="utf-8") as file:
                json_content = json.dumps(obj=info, indent=4)
                file.write(json_content)
        elif fmt == ExportFormat.CSV:
            with open(filename, mode="w", encoding="utf-8") as file:
                fieldnames = info[0].keys() if info else []
                writer = csv.DictWriter(file, fieldnames=fieldnames)
                writer.writeheader()
                writer.writerows(info)
        else:
            raise ValueError(f"Unsupported export format: {fmt}")

    def runs_info(
        self,
        status: Optional[Union[TaskStatusCode, str]] = None,
        select: Union[SelectMode, str] = SelectMode.DISTINCT,
    ) -> list:
        """
        Gathers the configuration and performance metrics for each run
        associated with the benchmark in a list, including computation cost and
        execution time.
        
        Args:
            status (Optional[Union[TaskStatusCode, str]]): The status of the
                tasks to include in the benchmarking results. Defaults to None,
                which includes all tasks.
            select (Union[SelectMode, str]): The data to include in
                the benchmarking results. Defaults to SelectMode.DISTINCT that
                includes only the parameters that vary between different runs.

        Returns:
            list: A list containing the configuration and performance 
                metrics for each run.
        """

        def get_task_input_params(task):
            input_filename = "input.json"
            input_dir_path = task.download_inputs(filenames=[input_filename])
            input_file_path = input_dir_path.joinpath(input_filename)
            with open(input_file_path, mode="r", encoding="utf-8") as file:
                return json.load(file)

        def select_distinct(info):
            attrs_lsts = defaultdict(list)
            for attrs in info:
                for attr, value in attrs.items():
                    attrs_lsts[attr].append(value)
            filtered = {attr for attr, values in attrs_lsts.items() \
                        if values and len(values) != values.count(values[0])}
            return [{attr: attrs[attr] for attr in filtered} for attrs in info]

        if isinstance(select, str):
            select = SelectMode[select.upper()]

        info = []
        tasks = self.get_tasks(status=status)
        for task in tasks:
            task_input_params = get_task_input_params(task)
            task_info = task.info
            task_machine_type = task_info.executer.vm_type \
                if task_info.executer else None
            task_time = task_info.time_metrics.computation_seconds.value
            task_cost = task_info.estimated_computation_cost
            info.append({
                Benchmark.InfoKey.TASK_ID: task_info.task_id,
                Benchmark.InfoKey.SIMULATOR: task_info.simulator,
                Benchmark.InfoKey.MACHINE_TYPE: task_machine_type,
                Benchmark.InfoKey.TIME: task_time,
                Benchmark.InfoKey.COST: task_cost,
                **task_input_params,
            })

        return select_distinct(info) \
            if select == SelectMode.DISTINCT \
            else info

    def terminate(self) -> Self:
        """
        Terminates all active machine groups associated with the
        benchmark.

        Returns:
            Self: The current instance for method chaining.
        """

        def _handle_suffix(executer):
            if executer.host_type == resources.machine_types.ProviderType.GCP:
                return "-".join(executer.vm_name.split("-")[:-1])
            return executer.vm_name

        machine_names = {
            _handle_suffix(task.info.executer)
            for task in self.get_tasks()
            if task.info.executer
        }

        for machine in resources.get():
            if machine.name not in machine_names:
                continue
            try:
                machine.terminate(verbose=False)
            except client.ApiException as api_exception:
                logging.warning(api_exception)

        return self



"""File Tracker module for connecting to a running task via WebRTC."""
import asyncio
import json
import uuid
import enum
import logging

aiortc_imported = True

try:
    import aiortc
except ImportError:
    aiortc_imported = False

# STUN/TURN server configuration
if aiortc_imported:
    ICE_SERVERS = [
        aiortc.RTCIceServer("stun:webrtc.inductiva.ai:3478"),
        aiortc.RTCIceServer("turn:webrtc.inductiva.ai:3478")
    ]

aiortc_logger = logging.getLogger("aioice")
aiortc_logger.setLevel(logging.WARNING)


class Operations(enum.Enum):
    LIST = "ls"
    TAIL = "tail"


class FileTracker:
    """File Tracker class for connecting to a running task via WebRTC."""

    def __init__(self):
        if not aiortc_imported:
            raise NotImplementedError("Feature not available for this version.")

        self.pc = aiortc.RTCPeerConnection(
            aiortc.RTCConfiguration(iceServers=ICE_SERVERS))
        self._message = None

    async def setup_channel(self, operation, **kwargs):
        channel = self.pc.createDataChannel("file_transfer")
        fut = asyncio.Future()

        @channel.on("open")
        def on_open():
            request = operation.value
            if kwargs:
                kwarg_str = map(str, kwargs.values())
                request += ":" + ",".join(kwarg_str)
            channel.send(request)

        @channel.on("message")
        def on_message(message):
            self._message = json.loads(message)
            fut.set_result(self._message)

        return fut

    async def connect_to_task(self, api, task_id):
        connection_id = str(uuid.uuid4())
        path_params = {"task_id": task_id}
        api.register_task(body={"sender_id": connection_id},
                          path_params=path_params)

        offer = await self.pc.createOffer()
        await self.pc.setLocalDescription(offer)

        api.offer_task(body={
            "sender_id": connection_id,
            "receiver_id": task_id,
            "type": "offer",
            "sdp": self.pc.localDescription.sdp
        },
                       path_params=path_params)

        resp = api.get_message(query_params={"client": connection_id},
                               path_params=path_params)

        if resp.response.status == 204:
            return False

        data = resp.body
        if data["type"] == "answer":
            await self.pc.setRemoteDescription(
                aiortc.RTCSessionDescription(sdp=data["sdp"],
                                             type=data["type"]))

        return True

    async def cleanup(self):
        await self.pc.close()
